#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun  4 15:47:58 2020

@author: ageiges
"""

import datatoolbox as dt
#%%
inventory = dt.find(source='IAMC')


replacementDict = {#'Capacity': 'Electricity|capacity',
                   'Capacity|Electricity' : 'Electricity|capacity',
                   'Heat_output_'   : 'Heat_output|',
                   'Losses_' : 'Losses|' ,
                   'Final_energy_demand_by_fuel|': 'Final_Energy|Total|',
                   'Final_energy' :'Final_Energy',
                   'Secondary_energy' : 'Secondary_Energy',
                   'Emissions|KyotoGHG' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|KYOTOGHG' : 'Emissions|KYOTOGHG',
                   'Emission|KYOTO_GHG_AR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto Gases|AR5-GWP100' : 'Emissions|KYOTOGHG_AR5',
                   'Emissions|KYOTOGHGAR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions_KYOTOGHGAR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto Gases|AR4-GWP10' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|KYOTOGHG_AR40' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions_KYOTOGHG_AR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto_Gases' : 'Emissions|KYOTOGHG',
                   'Emissions|Fuel|CO2' : 'Emissions|CO2|Fuel',
                   'Emissions|Fuel_CO2' : 'Emissions|CO2|Fuel',
                   'Exports_' : 'Exports|',
                   'population_total' : 'Population',
                   'population' : 'Population',
                   'gdp_ppp' : 'GDP|PPP',
                   'gdp_pp'  : 'GDP|PPP',
                   'GDP_PPP'  : 'GDP|PPP',
                   'GDP_MER'  : 'GDP|MER',
                   'Emissions_CH4' : 'Emissions|CH4',
                   'Emissions_CO2' : 'Emissions|CO2',
                   'Emissions_CO2_energy' : 'Emissions|CO2|Energy',
                   'Emissions|CO2_energy' : 'Emissions|CO2|Energy',
                   'Emissions|HFCS' : 'Emissions|HFCs',
                   'Emissions|PFCS' : 'Emissions|PFCs',
                   'Electricity_output'  : 'Electricity_generation',
                   'Electricity_output ' : 'Electricity_generation',
                   'Elect_Capacity'     : 'Electricity|capacity',  
                   'Electrical_capacity' : 'Electricity|capacity',
                   'Electrical|capacity' : 'Electricity|capacity',
                   'Electricity|capacity ': 'Electricity|capacity',
                   'Electricity_generation' : 'Electricity|generation',
                   'Electricity_genertation ' : 'Electricity|generation', 
                   'Elect_Generation' : 'Electricity|generation',
                   'Electricity_and_heat_generation' : 'Electricity&Heat|generation',
                   'Price_' : 'Price|',
                   'Primary_Energy_': 'Primary_Energy|',
                   'Primary_energy' : 'Primary_Energy',
                   'Production_' : 'Production|',
                   'Stock_changes_'  :'Stock_changes|'  ,
                   'Transfers_' : 'Transfers|',
                   'Total_PE_supply_' : 'Total_PE_supply|',
                   'Total_consumption_' : 'Total_consumption|',
                   'Emissions_er_capita' : 'Emissions_per_capita'}

for string, newString in replacementDict.items():
    mask = inventory.entity.str.contains(string)
    inventory.loc[mask, 'entity'] = inventory.loc[mask, 'entity'].apply(lambda x: x.replace(string,newString))
#%%
entityList = ['Electricity|generation|',
              'Electricity|capacity|',
              'Electricity&Heat|generation|',
              'Emissions|KYOTOGHG_AR4|',
              'Emissions|KYOTOGHG_AR5|',
              'Emissions|KYOTOGHG|',
              'Emissions|BC|',
              'Emissions|CO2|',
              'Emissions|CH4|',
              'Emissions|NH3|',
              'Emissions|N2O|',
              'Emissions|NF3|',
              'Emissions|NOx|',
              'Emissions|HFCs|',
              'Emissions|OC|',
              'Emissions|SF6|',
              'Emissions|PFCs|',
              'Emissions|VOC|',
              'Exports|',
              'Final_Energy|',
              'Investment|',
              'GDP|PPP|constant|',
              'GDP|PPP|current|',
              'GDP|MER|',
              'Heat_output|',
              'Secondary_Energy|',
              'Stock_changes|' ,
              'Transfers|',
              'Total_consumption|',
              'Population|',
              'Primary_Energy|',
              'Price|',
              'Production|']

inventory.category = None
for entity in entityList:
    mask = inventory.entity.str.startswith(entity)
    
    inventory.loc[mask, 'category'] = inventory.loc[mask, 'category'] + inventory.loc[mask, 'entity'].apply(lambda x: x.replace(entity,'')) 
    inventory.loc[mask, 'entity'] = entity[:-1]

#%%
entityList = list(inventory.entity.unique())
entityList.sort()
print(entityList)
print(len(entityList))


#%% ISSUES ??
if False:
    """
    'Total_consumption' : 'Final_energy'
    'Total_PE_supply'   : 'Primary_energy'
    'Primary_Energy_supply' : 'Primary_energy'
    'Production_'        : "?"
    'Transfers_'         : '?'
    'Stock_changes_'     
    Losses_
    'Energy_Demand'
    """
    
    #%%
dt.find(entity='Primary_Energy').source.unique()
#%%
dt.find(source='IEA_WEB_2019').entity.unique()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 11:07:40 2021

@author: ageiges
"""
import os
import shutil

#%% patch 0.4.5
def patch_045_update_personal_config(personal):
    
    MODULE_PATH = os.path.dirname(__file__)

    fin = open(os.path.join(MODULE_PATH, 'settings','personal.py'), 'r')
    lines = fin.readlines()
    fin.close()
    # os.makedirs(os.path.join(config.MODULE_PATH, 'settings'),exist_ok=True)
    fout = open(os.path.join(MODULE_PATH, 'settings','personal.py'), 'w')
    
    for line in lines:
        if line.endswith('\n'):
            outLine = line
        else:
            outLine = line + '\n'
            
        fout.write(outLine)
    
    # add it to old personal config
    outLine = 'AUTOLOAD_SOURCES = True'
    fout.write(outLine)
    fout.close()
    
    personal.AUTOLOAD_SOURCES = False
    
    return personal

def patch_047_move_config_file():
    from appdirs import user_data_dir
    appname = "datatoolbox"
    appauthor = "ageiges"
    CONFIG_DIR = user_data_dir(appname, appauthor)
    
    if os.path.isfile(os.path.join(os.path.dirname(__file__),'settings', 'personal.py')):
        print('Old configuration exists: APPLYING PATCH 47')
        
        if not os.path.exists(CONFIG_DIR):
            print('Creating new config folder')
            os.makedirs(CONFIG_DIR, exist_ok = True)
        
        print('Copying personal.py')
        shutil.copyfile(os.path.join(os.path.dirname(__file__), 'settings', 'personal.py'),
                        os.path.join(CONFIG_DIR,'personal.py'))
        print('removing old settings folder')
        # os.remove(os.path.join(os.path.dirname(__file__), 'settings', 'personal.py'))
        shutil.rmtree(os.path.join(os.path.dirname(__file__),'settings'))
if __name__ == '__main__':
    patch_047_move_config_file()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 18 15:48:29 2021

@author: ageiges
"""
# import datatoolbox as dt

from pymagicc.definitions import (
    MAGICC7_TO_MAGICC6_VARIABLES_MAPPING,
    convert_magicc7_to_openscm_variables,
    MAGICC7_EMISSIONS_UNITS)

ar4_gwp = {
    'CO2': 1,
    'CH4': 25,
    'N2O': 298,
    'CFC11': 4750,
    'CFC12': 10900,
    'CFC13': 14400,
    'CFC113': 6130,
    'CFC114': 10000,
    'CFC115': 7370,
    'HALON1301': 7140,
    'HALON1211': 1890,
    'HALON2402': 1640,
    'CCL4': 1400,
    'CH3BR': 5,
    'CH3CCL3': 146,
    'HCFC22': 1810,
    'HCFC123': 77,
    'HCFC124': 609,
    'HCFC141b': 725,
    'HCFC142b': 2310,
    'HCFC225ca': 122,
    'HCFC225cb': 595,
    'HFC23': 14800,
    'HFC32': 675,
    'HFC41': 92,
    'HFC125': 3500,
    'HFC134a': 1430,
    'HFC134': 1100,
    'HFC143a': 4470,
    'HFC143': 353,
    'HFC152': 53,
    'HFC152a': 124,
    'HFC161': 4,
    'HFC227ea': 3220,
    'HFC227a': 3220,
    'HFC236cb': 1210,
    'HFC236ea': 1330,
    'HFC236fa': 9810,
    'HFC245ca': 693,
    'HFC245fa': 1030,
    'HFC365mfc': 794,
    'HFC43-10': 1640,
    'SF6': 22800,
    'NF3': 17200,
    'CF4': 7390,
    'C2F6': 12200,
    'C3F8': 8830,
    'CC3F6': 9200,
    'CC4F8': 10300,
    'C4F10': 8860,
    'C5F12': 9160,
    'C6F14': 9300,
    'C7F16': 9300,
    'C10F18': 7190
}

def read_MAGICC6_scen_file(filepath):
    #%%
    import datatoolbox as dt
    from pymagicc.io import MAGICCData, read_cfg_file, NoReaderWriterError
    mdata = MAGICCData()
    mdata.read(filepath)
    mdata.df   

    tset = dt.TableSet()
    for variable in mdata.df.variable.unique():
        idx = mdata.df.variable.index[mdata.df.variable == variable]
        data = mdata.df.loc[idx,:]
        years = data.time
        values = data.value
        unit = data.unit[data.index[0]]
        if len(data.region.unique()) == 1:
            region = data.region[data.index[0]]
        meta = {'variable' : variable,
                'unit' : unit}
        edata = dt.Datatable([list(values)],
                             index=[region],
                             columns=years,
                             meta=meta
                             )
        if ('CO2' in variable) and ( "t C / yr" in unit):
            edata = edata.convert('Mt CO2 / yr')
            #print(edata)
        if ('N2O' in variable) and ( "t N2ON / yr" in unit):
            edata = edata.convert('Mt N2O / yr')
            #print(edata)
        
        tset[variable]  = edata
    return tset

#%%
def convert_to_CO2eq(tset):
    #%%
    
    for key in tset.keys():
        table = tset[key]
        
        unit = table.meta['unit']
        meta = table.meta.copy()
        for unit_to_test in ar4_gwp.keys():
            
            if unit_to_test in unit:
                print(f"{table.meta['variable']} -> COeq: {ar4_gwp[unit_to_test]}")
                table = table * ar4_gwp[unit_to_test]
                
                table.meta = meta
                table.meta['unit'] = table.meta['unit'].replace(unit_to_test, 'CO2eq')
                break
        else:
            print(f"Could not convert: {table.meta['variable']} -> COeq: 0")
            table = table * 0
            table.meta = meta
            table.meta['unit'] = table.meta['unit'].replace(unit_to_test, 'CO2eq')
            # table.meta['variable'] = table.meta['unit'].replace(unit_to_test, 'CO2eq')
            # print(f'Could not convert {table.meta["variable"]}')
        
        if 'kt' in table.meta['unit']:
            table = table.convert(table.meta['unit'].replace('kt', 'Mt'))
        tset[key] = table
    return tset
if __name__ == '__main__':
    #%%
    filepath = '/media/sf_Documents/python/cat_global_aggregation_2.0/output/211017_21_10_COP_0/magicc_results/MAGICCresults_CPPHIGH_CATFINAL_211017/QuantileCalculation/results/EQW_SPLIT_CPPHIGH_CATFINAL_211017.SCEN'
    dset = read_MAGICC6_scen_file(filepath)
    dset = convert_to_CO2eq(dset)
    df, meta  = dset.to_compact_long_format()
    
    df.loc[:,df.columns[3:]].sum().plot()
    
    filepath = '/media/sf_Documents/python/cat_global_aggregation_2.0/output/211017_21_10_COP_0/magicc_results/MAGICCresults_CPPLOW_CATFINAL_211017/QuantileCalculation/results/EQW_SPLIT_CPPLOW_CATFINAL_211017.SCEN'
    dset = read_MAGICC6_scen_file(filepath)
    dset = convert_to_CO2eq(dset)
    df, meta  = dset.to_compact_long_format()
    
    df.loc[:,df.columns[3:]].sum().plot()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun  8 09:10:13 2020

@author: ageiges
"""

import datatoolbox as dt

for source in dt.inventory().source.unique():
    try:
        IDList = list(dt.findExact(source= source).index)
        dt.util.zipExport(IDList, source + '.zip')
    except:
        print('source {} was not exported'.format(source))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  1 18:15:56 2020

@author: ageiges
"""

import datatoolbox as dt
import os
import numpy.testing as npt

from util_for_testing import df1, df_datetime

def test_datatable__csv_io():

    
    for df in df1, df_datetime:
    
        filePath = os.path.join('/tmp','test.csv')
        
        df.to_csv(filePath)
    
        df_copy = dt.data_structures.read_csv(filePath)
        
        obs = df_copy
        exp = df
        npt.assert_array_almost_equal(obs, exp, decimal = 6)
        
        os.remove(filePath)
    
def test_datatable_excel_io():

    
    
    for df in df1, df_datetime:
    
        filePath = os.path.join('/tmp','test.xlsx')
        
        df.to_excel(filePath)
    
        df_copy = dt.data_structures.read_excel(filePath)
        
        obs = df_copy
        exp = df
        npt.assert_array_almost_equal(obs, exp, decimal = 6)
        
        # os.remove(filePath)

if __name__== '__main__':
    test_datatable__csv_io()
    test_datatable_excel_io()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 12 09:14:25 2021

@author: ageiges
"""
import pyam
import pandas as pd
import xarray as xr
import regex
from .. import util

def _filter(data, level=None, regex=False, **filters):
    # filter by columns and list of values
    keep = True

    for field, pattern in filters.items():
        # treat `col=None` as no filter applied
        if pattern is None:
            continue

        if field not in data:
            raise ValueError(f'filter by `{field}` not supported')

        keep &= util.pattern_match(
            data[field], pattern, regex=regex
        )

    if level is not None:
        keep &= data['variable'].str.count(r"\|") == level

    table = pd.DataFrame(
        data if keep is True else data.loc[keep]
    )
    return table

def read_partial(filename, **filters):
    """
    Reads a pyam csv and returns a subset based on the filters given.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.
    filters : TYPE
        DESCRIPTION.

    Returns
    -------
    pyam.IamDataframe

    """
    df = pd.read_csv(filename)
    df = _filter(df, **filters)
    idf = pyam.IamDataFrame(df)
    
    return idf

def _pack_dimensions(index, 
                     meta = None,
                     **stacked_dims):
    packed_labels = {}
    packed_values = {}
    drop_levels = []
    
    for dim, levels in stacked_dims.items():
        labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in levels])
        packed_labels[dim] = labels_u = labels.unique()
        packed_values[dim] = pd.Index(labels_u.get_indexer(labels), name=dim)
        drop_levels.extend(levels)

    #meta 
    if meta is not None:
        labels = pd.MultiIndex.from_arrays([index.get_level_values(l) for l in meta.index.names])
        labels_u = labels.unique()
        packed_values['meta'] = pd.Index(labels_u.get_indexer(labels), name='meta')
        packed_labels['meta'] = pd.MultiIndex.from_arrays(meta.loc[labels_u,:].values.T, names = meta.columns)
    
    
    return (
        pd.MultiIndex.from_arrays(
            [index.get_level_values(l) for l in index.names.difference(drop_levels)] +
            list(packed_values.values())
        ),
        packed_labels
    )

def get_pathway_from_multiindex(index):
    return list(index.get_level_values('scenario') + '|' + index.get_level_values('model'))

def idf_to_xarray(s, stacked_dims=None):
    """
    Convert multiindex series or pyam dataframe to xarray
    """
    
    
    if isinstance(s, pyam.IamDataFrame):
         # meta
        meta_dims = dict(pathway=("model", "scenario"))
        index, labels = _pack_dimensions(s.meta.index, **meta_dims)
        x_meta = xr.DataArray(s.meta).rename({'dim_0':'pathway',
                                      'dim_1': 'meta'}).assign_coords(labels)
        # data
        s = s._data
    
    if stacked_dims is None:
        stacked_dims = dict(pathway=("model", "scenario"), varunit=("variable", "unit"))

    index, labels = _pack_dimensions(s.index, **stacked_dims)
    print(index)
    print(labels)
    return xr.DataArray.from_series(s.set_axis(index)).assign_coords(labels).assign_attrs({'iamc_meta' : x_meta})

def pd_long_to_xarray(s, **stacked_dims):
    """
    Convert multiindex series or pyam dataframe to xarray
    """
    
    
    if isinstance(s, pyam.IamDataFrame):
         # meta
        meta_dims = dict(pathway=("model", "scenario"))
        index, labels = _pack_dimensions(s.meta.index, **meta_dims)
        x_meta = xr.DataArray(s.meta).rename({'dim_0':'pathway',
                                      'dim_1': 'meta'}).assign_coords(labels)
        # data
        # s = s._data
    
    if not stacked_dims:
        stacked_dims = dict(pathway=("model", "scenario"), varunit=("variable", "unit"))

    index, labels = _pack_dimensions(s.index, **stacked_dims)
    
    return xr.DataArray.from_series(s.set_axis(index)).assign_coords(labels).assign_attrs({'iamc_meta' : x_meta})


# import datatoolbox as dt
# tbs = dt.getTables(dt.find().index[:10])
# idf = tbs.to_IamDataFrame()
# xda = idf_to_xarray(idf)
# labels
# tbs.xda

def compute_ghg_emissions(idf,
                          aggregated_variable = 'Emissions|Kyoto Gases',
                          variables = ['Emissions|CO2', 'Emissions|CH4', 'Emissions|N2O','Emissions|F-gases'],
                          append = True):
    
    org_timeseries = idf.timeseries()
    ghg_data = (idf.filter(variable=variables).
                convert_unit('Mt CH4/yr', 'Mt CO2-equiv/yr', factor =25).
                convert_unit('kt N2O/yr', 'Mt CO2-equiv/yr', factor =0.298).
                convert_unit('Mt CO2/yr', 'Mt CO2-equiv/yr', factor =1)
        )
    
    if append:
        ghg_data.aggregate(aggregated_variable, components=variables,append=True)
        ghg_timeseries = ghg_data.timeseries()
        idx_to_add = ghg_timeseries.index.difference(org_timeseries.index)
        return pyam.IamDataFrame(org_timeseries.append(ghg_timeseries.loc[idx_to_add,:]))
    else:
        return ghg_data.aggregate(aggregated_variable, components=variables)
def compute_ghg_excluding_landuse(idf):
    comp_data = (idf.filter(variable =['Emissions|Kyoto Gases', 'Emissions|CO2|Land Use']).
           convert_unit('Mt CO2/yr', 'Mt CO2eq/yr', factor =1).
           convert_unit('Mt CO2-equiv/yr', 'Mt CO2eq/yr', factor =1)
           )
    comp_data.subtract('Emissions|Kyoto Gases', 'Emissions|CO2|Land Use', name='Emissions|Kyoto excl LULUCF', append=True, ignore_units=True)
    return comp_data
    
def complete_world_emission(idf, 
                            
                       reg_mapping= {'World' : ['ASIA', 'LAM', 'MAF', 'OECD90', 'REF']}):
    #%%
    # idf = native_AR5_data.copy()
    org_timeseries = idf.timeseries()
    agg_data = list()
    for variable in variables:
        for region, sub_regions in reg_mapping.items():
            agg_data.append(idf.aggregate_region(variable, region='World'))
    agg_data = pyam.concat(agg_data)
    agg_timeseries = agg_data.timeseries()
    idx_to_add = agg_timeseries.index.difference(org_timeseries.index)
    agg_data = pyam.IamDataFrame(org_timeseries.append(agg_timeseries.loc[idx_to_add,:]))
    ghg_data = (agg_data.
                convert_unit('Mt CH4/yr', 'Mt CO2-equiv/yr', factor =25).
                convert_unit('kt N2O/yr', 'Mt CO2-equiv/yr', factor =0.298).
                convert_unit('Mt CO2/yr', 'Mt CO2-equiv/yr', factor =1)
        )
    ghg_data = ghg_data.append(ghg_data.aggregate('Emissions|Kyoto Gases', components=variables))
    ghg_timeseries = ghg_data.timeseries()
    idx_to_add = ghg_timeseries.index.difference(org_timeseries.index)
    return pyam.IamDataFrame(org_timeseries.append(ghg_timeseries.loc[idx_to_add,:]))
    
    # return ghg_data.append(agg_data).append(idf)
#%%
# tbs = dt.getTables(dt.find(entity='Emissions|CO2').index[:10])
# xds = xr.Dataset()
# stacked_dims= {'pathway': ('model', 'scenario')}
# for var in tbs.variables():
#     idf = tbs.filter(variable=var).to_LongTable().set_index(['variable','scenario','model', 'region'])
#     xds[var] = pd_long_to_xarray(idf, **stacked_dims)



##!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Core contains some very basic functions that are used within the package
in various locations and tools.

@author: Andreas Geiges
"""

import time
import pint

import re

import numpy as np
import pandas as pd
from datatoolbox import config
from datatoolbox import naming_convention

tt = time.time()

#%% Pint unit handling
gases = {"CO2eq":"carbon_dioxide_equivalent",
         "CO2e" : "CO2eq",
         "NO2" : "NO2",
         'PM25' : 'PM25'}

from openscm_units import unit_registry as ur


try:
    ur._add_gases(gases)
    ur.define('fraction = [] = frac')
    ur.define('percent = 1e-2 frac = pct')
    ur.define('ppm = 1e-6 fraction')
    ur.define('sqkm = km * km')
    ur.define('none = dimensionless')


    ur.load_definitions(config.PATH_PINT_DEFINITIONS)
except pint.errors.DefinitionSyntaxError:
    # avoid double import of units defintions
    pass
    
import pint

# 
#     from .tools import xarray as _xr
#     to_XDataSet = _xr.to_XDataSet
#     to_XDataArray = _xr.to_XDataArray
if config.AVAILABLE_XARRAY:   
    import pint_xarray
    pint_xarray.accessors.setup_registry(ur)
    pint_xarray.unit_registry = ur
    pint_xarray.accessors.default_registry = ur

#%%

c = pint.Context('GWP_AR5')

CO2EQ_LIST = ['CO2eq', 'CO2e',]

AR4GWPDict = {'CH4': 25,
              'HFC': 1430,
              'N2O' : 298,
              'SF6' : 22800,
              'PFC' : 7390,
              'CO2' : 1}


LOG = dict()
LOG['tableIDs'] = list()

ur.add_context(c)

def is_known_entity(variable):
    entity_matches = list()
    for entity in naming_convention.entities:
        if variable.startswith(entity):
            return True
    
    return False
    
def _split_variable(metaDict):
    """
    Split variable into a known entity (see naming_converntion.py) and a 
    category.
    
    Parameters
    ----------
    metaDict : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    
    #Find entity
    entity_matches = list()
    for entity in naming_convention.entities:
        if metaDict['variable'].startswith(entity):
            entity_matches.append(entity)
    
    if len(entity_matches)> 0:
        longest_matchg = max(entity_matches, key=len)
        metaDict['entity'] = longest_matchg
    
    else:
        if config.DEBUG:
            print(f'Warning: Entity could not be derived from {metaDict["variable"]}')
            
        # exit here
        return metaDict
        
        
    #derive or check category
    if 'category' not in metaDict.keys():
        metaDict['category'] = metaDict['variable'].replace(longest_matchg,'').strip('|')
    else:
        if metaDict['category'] != metaDict['variable'].replace(longest_matchg,'').strip('|'):
            print('Warming current category not fitting derived category, please review')
    return metaDict



def _update_meta(metaDict):
    """
    Private funcion to update the meta data of a datatable
    

    Parameters
    ----------
    metaDict : dict
        new data to overwrite.

    Returns
    -------
    metaDict : dict

    """
    if 'entity' not in metaDict.keys():
        
        metaDict = _split_variable(metaDict)
        
        
    for key in list(metaDict.keys()):
        if (metaDict[key] is np.nan) or metaDict[key] == '':
            if key != 'unit':
                del metaDict[key]
            
    for id_field in config.ID_FIELDS:
        fieldList = [ metaDict[key] for key in config.SUB_FIELDS[id_field] if key in  metaDict.keys()]
        if len(fieldList)>0:
            new_value = config.SUB_SEP[id_field].join([str(x) for x in fieldList]).strip('|')
            if config.DEBUG and id_field in metaDict.keys() and metaDict[id_field] != new_value:
                print(f'Warning: {id_field} will be overritten {metaDict[id_field]} -> {new_value}')
            metaDict[id_field] =  new_value
    
    return metaDict

def _fix_filename(name, max_length=255):
    """
    Replace invalid characters on Linux/Windows/MacOS with underscores.
    List from https://stackoverflow.com/a/31976060/819417
    Trailing spaces & periods are ignored on Windows.
    >>> fix_filename("  COM1  ")
    '_ COM1 _'
    >>> fix_filename("COM10")
    'COM10'
    >>> fix_filename("COM1,")
    'COM1,'
    >>> fix_filename("COM1.txt")
    '_.txt'
    >>> all('_' == fix_filename(chr(i)) for i in list(range(32)))
    True
    """
    return re.sub(r'[/\\:<>"?*\0-\x1f]|^(AUX|COM[1-9]|CON|LPT[1-9]|NUL|PRN)(?![^.])|^\s|[\s.]$', "_", name[:max_length], flags=re.IGNORECASE)

def _validate_unit(table):
    
    try:
        getUnit(table.meta['unit'])
        
        return True
    except:
        return False

def generate_table_file_name(ID):
    ID_for_filename = _fix_filename(ID) 
    ID_for_filename = ID.replace('|','-').replace('/','-')
    return ID_for_filename + '.csv'

def _createDatabaseID(metaDict):
    ID =config.ID_SEPARATOR.join(
        [metaDict[key] for key in config.ID_FIELDS]
        )
    # ID = _fix_filename(ID)
    return ID

def csv_writer(filename, 
               dataframe,
               meta,
               index=0):
    fid = open(filename,'w', encoding='utf-8')
    fid.write(config.META_DECLARATION)
    
    for key, value in sorted(meta.items()):
#            if key == 'unit':
#                value = str(value.u)
        fid.write(key + ',' + str(value) + '\n')
    
    fid.write(config.DATA_DECLARATION)
    if index == 0:
        dataframe.to_csv(fid, sep = ',')
    elif index is None:
        dataframe.to_csv(fid,index=None,sep = ';')
    fid.close()

def excel_writer(writer,
                 dataframe,
                 meta,
                 sheet_name="Sheet1",
                 index=False,
                 engine=None):
    
    if isinstance(writer, pd.ExcelWriter):
        need_save = False
    else:
        writer = pd.ExcelWriter(pd.io.common.stringify_path(writer),  engine=engine)
        need_save = True
        
    
    metaSeries= pd.Series(data=[''] + list(meta.values()) + [''],
                              index=['###META###'] + list(meta.keys()) + ['###DATA###'])
        
    metaSeries.to_excel(writer, sheet_name=sheet_name, header=None, columns=None)
    pd.DataFrame(dataframe).to_excel(writer, sheet_name= sheet_name, index= index, startrow=len(metaSeries))
    
    if need_save:
        writer.save()
            

def osIsWindows():
    """
    Checkes if operating system is windows based
    

    Returns
    -------
    bool

    """
    if (config.OS == 'win32') | (config.OS == "Windows"):
        return True
    else:
        return False

def getUnit(string):
    """
    Return the pint unit for a given unit string. Compared to the original
    pint this functions replaces special characters $ € and % by a string 
    reprentation.

    Parameters
    ----------
    string : str
        unit string (e.g. "km / s" or "€  / capita")

    Returns
    -------
    unit : pint unit
    """
    # if not isinstance(string, str):
    #     string = str(string)
    if string is None:
        string = ''
    else:
        string = string.replace('$', 'USD').replace('€','EUR').replace('%','percent')
    return ur(string)

def getUnitWindows(string):
    """
    Equivalent version of getUnit  but adapted for windows system.
    
    Parameters
    ----------
    string : str
        unit string (e.g. "km / s" or "€  / capita")

    Returns
    -------
    unit : pint unit

    """
    if string is None:
        string = ''
    else:
        string = string.replace('$', 'USD').replace('€','EUR').replace('%','percent').replace('Â','')
    return ur(string)

# re-defintion of getUnit function for windows users
if osIsWindows():
    getUnit = getUnitWindows


def getTimeString():
    """
    Return formated time string.

    Returns
    -------
    time string : str

    """
    return time.strftime("%Y/%m/%d-%I:%M:%S")

def getDateString():
    """
    Return formated date string.

    Returns
    -------
    date string : str

    """
    return time.strftime("%Y_%m_%d")

def conversionFactor(unitFrom, unitTo, context =None):
    """
    Return the conversion factor from one unit to another.

    Parameters
    ----------
    unitFrom : str
        Original unit to convert.
    unitTo : str
        Unit to which it original unit should be converted.
    context : str, optional
        For some conversions, a specifice context is needed. Currently, only
        GWPAR4 is implemented. The default is None.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """ 
    if context is None:
        return getUnit(unitFrom).to(getUnit(unitTo)).m
    elif context == 'GWPAR4':
        
        return _AR4_conversionFactor(unitFrom, unitTo)
        
    else:
        raise(BaseException('unkown context'))


def _findGases(string, candidateList):
    hits = list()
    for key in candidateList:
        if key in string:
            hits.append(key)
            string = string.replace(key,'')
    return hits

def _AR4_conversionFactor(unitFrom, unitTo):
#    weirdSet = set(['CO2','CO','VOC', 'OC'])
    
    # look if unitTo is CO2eq -> conversion into co2 equivalent
    co2eqkeys = _findGases(unitTo, CO2EQ_LIST)        
    gasesToconvert = _findGases(unitFrom, list(AR4GWPDict))
    
    assert len(co2eqkeys) == 1 and len(gasesToconvert) == 1
    co2Key = co2eqkeys[0]
    gasKey = gasesToconvert[0]
        
    if config.DEBUG:    
        print('Converting from {} to {} using GWP AR4'.format(gasKey,co2Key))
        
    
    unitFrom = unitFrom.replace(gasKey, co2Key)
    conversFactor = getUnit(unitFrom).to(unitTo).m
    co2eq_factor = AR4GWPDict[gasKey]
    factor = conversFactor * co2eq_factor
    return factor


#%%
def get_dimension_extend(table_iterable, dimensions):
    """
    This functions assesses the the unique extend for various dimensions
    given a set of datatables
    """
    fullIdx = dict()
    # for dim in dimensions:
    #     fullIdx[dim] = set()

    for table in table_iterable:
    
#        for metaKey, metaValue in table.meta.items():
#            if metaKey not in metaDict.keys():
#                metaDict[metaKey] = set([metaValue])
#            else:
#                metaDict[metaKey].add(metaValue)
            
        for dim in dimensions:
            
            if dim not in fullIdx.keys():
                fullIdx[dim] = set()
                
            if dim == 'region':
                fullIdx[dim] = fullIdx[dim].union(table.index)
            elif dim == 'time':
                fullIdx[dim] = fullIdx[dim].union(table.columns)
            elif dim in table.meta.keys():
                fullIdx[dim].add(table.meta[dim])
            else:
                raise(BaseException('Dimension not available'))

    dimSize = [len(fullIdx[x]) for x in dimensions]
    dimList = [sorted(list(fullIdx[x])) for x in dimensions]
    
    return dimSize, dimList
    

def get_meta_collection(table_iterable, dimensions):
    """

    Parameters
    ----------
    table_iterable : list of tables
        DESCRIPTION.
    dimensions : list of dimentions
        DESCRIPTION.

    Returns
    -------
    metaCollection : TYPE
        DESCRIPTION.

    """

    
    metaCollection = dict()
    for table in table_iterable:
        
        for key in table.meta.keys():
            if key in dimensions  or key == 'ID':
                continue
            if key not in metaCollection.keys():
                metaCollection[key] = set()
                
            metaCollection[key].add(table.meta[key])
    
    return metaCollection
    
if config.DEBUG:
    print('core loaded in {:2.4f} seconds'.format(time.time()-tt))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 18 09:03:50 2021

@author: ageiges
"""

"""

This tutorial covers the datatoolbox inventory and  two search functions 
for datatables

1) Inventory
2) findp "pattern"
3) findc "conains"

"""


import datatoolbox as dt

#%% tutorial data
"""
This tutoral works based on the PRIMAP emission data set. To work with this tutorial,
please import the data.
"""

# if 'PRIMAP'
    dt.import_new_source_from_remote('PRIMAP_2019')
    
    #%% Inventory
    """
    The inventory is a pandas dataframe containing all available tables, each in a row.
    The index has the tableIDs and comes with the following colums : 
    ['variable', 'entity', 'category', 'pathway', 'scenario', 'model', 'source', 'source_name', 'source_year', 'unit']
    
    All coulumns can be used for the later search requests.
    """
    dt.inventory().head()
    
    
    #%% Seach using patterns - findp
    """
    Defauls is a shell style type of search suitable to be used for categories.
    """
    
    dt.findp(source = 'PRIMAP_2019')
    
    # One star (*) is giving all results for only the following level (speparated by "|") 
    dt.findp(variable='Emissions|*',
             source = 'PRIMAP_2019') # "Emissions|KYOTOGHG_AR4|Marine" will be excluded

# Two stars (**) allows for any patter for the complete rest of the string
dt.findp(variable='Emissions|**',
         source = 'PRIMAP_2019') # "Emissions|KYOTOGHG_AR4|Marine" will be included

# One star (*) is giving all results for only the following level (speparated by "|") 
dt.findp(variable='Emissions|*|IPCM0EL',
         source = 'PRIMAP_2019') # "All kinds of IPCM0EL emissions, allows for different gases.

# several patters are Or'd together
dt.findp(variable=['Emissions|CO2|*', 'Emissions|CH4|*'],
         source = 'PRIMAP_2019')

# Example finding GHG emission for certain pathways

dt.findp(source = 'PRIMAP_2019').pathway.unique()

inv = dt.findp(entity='Emissions|KYOTOGHG_AR4', 
               pathway=['Historic|country_reported', 'Historic|third_party'],
               source = 'PRIMAP_2019')

#%% Seach using contains - findc
"""
Findc uses the panfas "contains" logic to filter the inventoy. Any findc search for a "string_pattern
can be reproduces by findp(**string_pattern**)
"""

dt.findc(variable='KYOTO', source = 'PRIMAP_2019')
#  equals
dt.findp(variable='**KYOTO**', source = 'PRIMAP_2019')

# Example finding Historic CO2 emission data sets

res = dt.find(entity='Emissions|CO2',
        scenario = 'Historic')
print(res.source.unique())
print(res.variable.unique())


res = dt.findp(entity='Emissions|CO2',
        scenario = 'Historic**')
print(res.source.unique())
print(res.variable.unique())

#%% conversions to other data structures

# load as pyam dataframe
idf = res.as_pyam()

#load as xarray 
xda = res.as_xarray()

#both allow to use the native region defintions
idf = res.as_pyam(native_regions=True)
xda = res.as_xarray(native_regions=True)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  8 11:25:15 2019

@author: Andreas Geiges
"""

import os, sys
import networkx as nx
from functools import reduce
from operator import and_

import pandas as pd
import numpy as np
from typing import Union, Iterable

import logging

import datatoolbox as dt
from datatoolbox import mapping as mapp
from datatoolbox import core
from datatoolbox import config
from datatoolbox import data_structures
from datatoolbox.greenhouse_gas_database import GreenhouseGasTable 
import matplotlib.pylab as plt

import tqdm

import deprecated as dp
#from .tools import kaya_idendentiy_decomposition

logger = logging.getLogger(__name__)

try:
    from hdx.location.country import Country
    
    def getCountryISO(string):
        #print(string)
        try:
            string = string.replace('*','')
            results = Country.get_iso3_country_code_fuzzy(string)
            if len(results)> 0:
                return results[0]
            else:
                return None
        except:
            print('error for: ' + string)
        
except:
    def getCountryISO(string):
        print("the package hdx is not installed, thus this function is not available")
        print('use: "pip install hdx-python-country" to install')
        return None



# colorGenerator = dp.deprecated(_plt.colorGenerator, 
#                                reason="please import from matplotlib toolkid", 
#                                version='0.4.0')
# savefig = dp.deprecated(_plt.savefig,
#                                reason="please import from matplotlib toolkid", 
#                                version='0.4.0')

#%%
from .tools import html as _html

def export_to_pdf( dataframe, fileName):
    html = _html.export_to_html(dataframe)
    import pdfkit
    pdfkit.from_string(html, fileName)


GHG_data = GreenhouseGasTable()

def diff_eps(df1, df2, eps = 1e-6):
    """
    Identify differences between two Datatables higher than a defined treshold.
    
    Returns a pandas dataframe with all changes
    """
    assert (df1.columns == df2.columns).all(), \
        "DataFrame column names are different"
    if any(df1.dtypes != df2.dtypes):
        "Data Types are different, trying to convert"
        df2 = df2.astype(df1.dtypes)
    if df1.equals(df2):
        return None
    else:
        # need to account for np.nan != np.nan returning True
        diff_mask = ((df1 - df2).abs()>eps) & ~(((df1.isnull()) & df2.isnull()).isnull())
        ne_stacked = diff_mask.stack()
        changed = ne_stacked[ne_stacked]
        changed.index.names = ['id', 'col']
        difference_locations = np.where(diff_mask)
        changed_from = df1.values[difference_locations]
        changed_to = df2.values[difference_locations]
        return pd.DataFrame({'from': changed_from, 'to': changed_to},
                            index=changed.index)
        
def diff(df1, df2):
    """
    Identify differences between two Datatables.
    
    Returns a pandas dataframe with all changes
    """
    assert (df1.columns == df2.columns).all(), \
        "DataFrame column names are different"
    if any(df1.dtypes != df2.dtypes):
        "Data Types are different, trying to convert"
        df2 = df2.astype(df1.dtypes)
    if df1.equals(df2):
        return None
    else:
        # need to account for np.nan != np.nan returning True
        diff_mask = (df1 != df2) & ~(df1.isnull() & df2.isnull())
        ne_stacked = diff_mask.stack()
        changed = ne_stacked[ne_stacked]
        changed.index.names = ['id', 'col']
        difference_locations = np.where(diff_mask)
        changed_from = df1.values[difference_locations]
        changed_to = df2.values[difference_locations]
        return pd.DataFrame({'from': changed_from, 'to': changed_to},
                            index=changed.index)        



def isUnit(unit):
    try:
        core.getUnit(unit)
    except:
        return False
    else:
        return True

def scatterIndicatorComparison(tableX, tableY):
    timeCol = list(set(tableY.columns).intersection(set(tableX.columns)))
    for ISOcode in tableX.index:
        coName= mapp.countries.codes.name.loc[ISOcode]
    #    
    #    p = plt.plot(tableX.loc[ISOcode,timeCol4], tableY.loc[ISOcode,timeCol4],linewidth=.2)
    #    color = p[0].get_color()
    #    p = plt.scatter(tableX.loc[ISOcode,timeCol], tableY.loc[ISOcode,timeCol], facecolors=color, edgecolors=color)
    #    plt.scatter(tableX.loc[ISOcode,timeCol2], tableY.loc[ISOcode,timeCol2], s=40, facecolors='none', edgecolors=color)
    #    plt.scatter(tableX.loc[ISOcode,timeCol3], tableY.loc[ISOcode,timeCol3], s=40, facecolors='none', edgecolors=color)
    #    
        p = plt.scatter(tableX.loc[ISOcode,timeCol], tableY.loc[ISOcode,timeCol], s = 100)
        
        plt.text(tableX.loc[ISOcode,timeCol]+.1, tableY.loc[ISOcode,timeCol]+.2, coName)
    plt.xlabel(tableX.ID)
    plt.ylabel(tableY.ID)
    
    plt.xlim([0,60])
    plt.ylim([0,20])
    plt.tight_layout()

#def harmonize_IPCC(startYear, endYear, startValue=None, endValue=None):
#    
#    assert (startValue is not None) or (endValue is not None)
#    
#    if endValue is None:
#        
#        
#        e
    
def cleanDataTable(dataTable):
    # ensure valid ISO or region IDs
    # invalidSpatialIDs = dataTable.index.difference(mapp.getValidSpatialIDs())
    # if not invalidSpatialIDs.empty:
    #     print(
    #         f"Removing unknown regions from dataTable {dataTable.meta.get('ID', 'unnamed')}:",
    #         ", ".join(invalidSpatialIDs)
    #     )
    #     dataTable = dataTable.loc[dataTable.index.difference(invalidSpatialIDs)]
    if 'standard_region' in dataTable.columns:
        dataTable = dataTable.reset_index('region').set_index(['region', 'standard_region'])
    dataTable = dataTable.dropna(how="all", axis=1).dropna(how="all", axis=0).astype(float)

    # clean meta data
    keysToDelete = list()
    for key in dataTable.meta.keys():
        if np.any(pd.isna(dataTable.meta[key])):
            if key not in config.ID_FIELDS:
                keysToDelete.append(key)
            else:
                dataTable.meta[key] = ''
    for key in keysToDelete:
        del dataTable.meta[key]

    # ensure time colums to be integer
    dataTable.columns = dataTable.columns.astype(int)
    
    dataTable = dataTable.loc[:, dataTable.columns.sort_values()]
    return dataTable




    
    
def identifyCountry(string):
    
    #numeric ISO code
    try:
        numISO = float(string)
        mask = numISO == mapp.countries.codes['numISO']
        if mask.any():
            return mapp.countries.index[mask][0]
    except:
        pass
    
    if len(str(string)) == 2:
        mask = str(string).upper() == mapp.countries.codes['alpha2']
        if mask.any():
            return mapp.countries.codes.index[mask][0]
    
    if len(str(string)) == 3:
         
        if str(string).upper() in mapp.countries.codes.index:
            return string.upper()
    
    try: 
        coISO = dt.getCountryISO(string)
        return coISO
    except:
        print(f'not matching country found for {string}')    
        return None
            
def convertIndexToISO(table, iso_type='alpha3'):
    """
    Convert index of a dataframe into iso codes.

    Parameters
    ----------
    table : pandas.Dataframe or dt.DataTable
        Index of thos table consists of country strings.
    iso : TYPE, optional
        Either 'alpha3', alpha2 or numISO. The default is 'alpha3'.

    Returns
    -------
    table :  pandas.Dataframe or dt.DataTable
        Return old dataframe with new iso index.

    """
    replaceDict = dict()
    
    for idx in table.index:
        iso = identifyCountry(idx)
        if iso is not None:
            replaceDict[idx] = iso
    table.index = table.index.map(replaceDict)           
    table = table.loc[~table.index.isna(),:]

    if iso_type == 'alpha2':
        table.index = mapp.countries.codes.loc[table.index,'alpha2']
    elif iso_type == 'numISO':
        table.index = mapp.countries.codes.loc[table.index,'numISO'].astype(int)
    return table

def addCountryNames(table, as_index=False):
    names = list()
    for idx in table.index:
        if idx in mapp.countries.codes.index:
            names.append(mapp.countries.codes.loc[idx,'name'])
        else:
            names.append(idx)
    if as_index:
        table.index = names    
    else:
        table.loc[:,'country_name'] = names
    return table




def update_source_from_file(fileName, message=None):
    sourceData = pd.read_csv(fileName)
    for index in sourceData.index:
        dt.core.DB._addNewSource(sourceData.loc[index,:].to_dict())
    
def update_DB_from_folder(folderToRead, message=None):
    fileList = os.listdir(folderToRead)
    fileList = [file for file in fileList if '.csv' in file[-5:].lower()]

    tablesToUpdate = dict()

    for file in fileList:
        table = dt.read_csv(os.path.join(folderToRead, file))
        source = table.meta['source']
        if source in tablesToUpdate.keys():


            tablesToUpdate[source].append(table)
        else:
            tablesToUpdate[source] = [table]
    if message is None:

        message = 'External data added from external source by ' + config.CRUNCHER
    
    for source in tablesToUpdate.keys():
        sourceMetaDict = dict()
        sourceMetaDict['SOURCE_ID']= source
        dt.commitTables(tablesToUpdate[source], 
                        message = message, 
                        sourceMetaDict = sourceMetaDict, 
                        append_data=True, 
                        update=True)


def update_DB_from_folderV3(folderToRead, message=None, cleanTables=True):
    import math
    fileList = os.listdir(folderToRead)
    fileList = [file for file in fileList if '.csv' in file[-5:].lower()]

    

    filesPerCommit = 5000
    nCommits = math.ceil((len(fileList))/filesPerCommit)
    for nCommit in range(nCommits):
        tablesToUpdate = dict()
        for file in fileList[nCommit*filesPerCommit:(nCommit+1)*filesPerCommit]:

            table = dt.read_csv(os.path.join(folderToRead, file))
            source = table.meta['source']
            
#            if not 'Emissions|CO2|Industrial' in table.ID:
#                continue
            if source in tablesToUpdate.keys():
    
    
                tablesToUpdate[source].append(table)
            else:
                tablesToUpdate[source] = [table]
        
        if message is None:
    
            message = 'External data added from external source by ' + config.CRUNCHER + '{}/{}'.format(nCommit,nCommits)
    
        for source in tablesToUpdate.keys():
    
            tablesToUpdate[source] = metaV2_to_meta_V3(tablesToUpdate[source])
#        return tablesToUpdate
    
        for source in tablesToUpdate.keys():
            sourceMetaDict = dict()
            sourceMetaDict['SOURCE_ID']= source
            core.DB.commitTables(tablesToUpdate[source], 
                            message = message,
                            sourceMetaDict = sourceMetaDict, 
                            cleanTables=cleanTables,
                            update=True)

#def update_DB_from_folderV3(folderToRead, message=None, cleanTables=True):
#    import math
#    fileList = os.listdir(folderToRead)
#    fileList = [file for file in fileList if '.csv' in file[-5:].lower()]
#
#    
#
##    filesPerCommit = 5000
##    nCommits = math.ceil((len(fileList))/filesPerCommit)
##    for nCommit in range(nCommits):
#    tablesToUpdate = dict()
#    for file in fileList:
#
#        table = dt.read_csv(os.path.join(folderToRead, file))
#        source = table.meta['source']
#        
#        if not 'Emissions|CO2|Industrial' in table.ID:
#                continue
#        if source in tablesToUpdate.keys():
#
#
#            tablesToUpdate[source].append(table)
#        else:
#            tablesToUpdate[source] = [table]
#    
##    if message is None:
##
##        message = 'External data added from external source by ' + config.CRUNCHER + '{}/{}'.format(nCommit,nCommits)
#
#    for source in tablesToUpdate.keys():
#
#        tablesToUpdate[source] = metaV2_to_meta_V3(tablesToUpdate[source])
#    return tablesToUpdate
    
#        for source in tablesToUpdate.keys():
#            sourceMetaDict = dict()
#            sourceMetaDict['SOURCE_ID']= source
#            core.DB.commitTables(tablesToUpdate[source], 
#                            message = message,
#                            sourceMetaDict = sourceMetaDict, 
#                            cleanTables=cleanTables)

def metaV2_to_meta_V3(tableSet):
    replacementDict = {#'Capacity': 'Electricity|capacity',
                   'Capacity|Electricity' : 'Electricity|capacity',
                   'Heat_output_'   : 'Heat_output|',
                   'Losses_' : 'Losses|' ,
                   'Final_energy_demand_by_fuel|': 'Final_Energy|Total|',
                   'Final_energy' :'Final_Energy',
                   'Secondary_energy' : 'Secondary_Energy',
                   'Emissions|KyotoGHG' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|KYOTOGHG' : 'Emissions|KYOTOGHG',
                   'Emission|KYOTO_GHG_AR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emission|KYOTOGHG'      : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto Gases|AR5-GWP100' : 'Emissions|KYOTOGHG_AR5',
                   'Emission|KYOTO_GHG_AR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions_KYOTOGHGAR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto Gases|AR4-GWP10' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|KYOTOGHG_AR40' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions_KYOTOGHG_AR4' : 'Emissions|KYOTOGHG_AR4',
                   'Emissions|Kyoto_Gases' : 'Emissions|KYOTOGHG',
                   'Emissions|Fuel|CO2' : 'Emissions|CO2|Fuel',
                   'Emissions|Fuel_CO2' : 'Emissions|CO2|Fuel',
                   'Exports_' : 'Exports|',
                   'population_total' : 'Population',
                   'population' : 'Population',
                   'gdp_ppp' : 'GDP|PPP',
                   'gdp_pp'  : 'GDP|PPP',
                   'GDP_PPP'  : 'GDP|PPP',
                   'GDP_MER'  : 'GDP|MER',
                   'Emissions_CH4' : 'Emissions|CH4',
                   'Emissions_CO2' : 'Emissions|CO2',
                   'Emissions_CO2_energy' : 'Emissions|CO2|Energy',
                   'Emissions|CO2_energy' : 'Emissions|CO2|Energy',
                   'Emissions|HFCS' : 'Emissions|HFCs',
                   'Emissions|PFCS' : 'Emissions|PFCs',
                   'Electricity_output'  : 'Electricity_generation',
                   'Electricity_output ' : 'Electricity_generation',
                   'Elect_Capacity'     : 'Electricity_capacity',  
                   'Electrical_capacity' : 'Electricity_capacity',
                   'Electrical|capacity' : 'Electricity_capacity',
                   'Electricity|capacity ': 'Electricity_capacity',
                   'Electricity»generation' : 'Electricity_generation',
                   'Electricity»genertation ' : 'Electricity_generation', 
                   'Elect_Generation' : 'Electricity_generation',
                   'Electricity_and_heat_generation' : 'Electricity&Heat_generation',
                   'Price_' : 'Price|',
                   'Primary_Energy_': 'Primary_Energy|',
                   'Primary_energy' : 'Primary_Energy',
                   'Production_' : 'Production|',
                   'Stock_changes_'  :'Stock_changes|'  ,
                   'Transfers_' : 'Transfers|',
                   'Total_PE_supply_' : 'Total_PE_supply|',
                   'Total_consumption_' : 'Total_consumption|',
                   'Emissions_per_capita' : 'Emissions_per_capita'}

    entityList = ['Electricity|generation|',
                  'Electricity|capacity|',
                  'Electricity&Heat|generation|',
                  'Emissions|KYOTOGHG_AR4|',
                  'Emissions|KYOTOGHG_AR5|',
                  'Emissions|KYOTOGHG|',
                  'Emissions|BC|',
                  'Emissions|CO2|',
                  'Emissions|CH4|',
                  'Emissions|NH3|',
                  'Emissions|N2O|',
                  'Emissions|NF3|',
                  'Emissions|NOx|',
                  'Emissions|HFCs|',
                  'Emissions|OC|',
                  'Emissions|SF6|',
                  'Emissions|PFCs|',
                  'Emissions|VOC|',
                  'Exports|',
                  'Final_Energy|',
                  'Investment|',
                  'GDP|PPP|constant|',
                  'GDP|PPP|current|',
                  'GDP|MER|',
                  'Heat_output|',
                  'Secondary_Energy|',
                  'Stock_changes|' ,
                  'Transfers|',
                  'Total_consumption|',
                  'Population|',
                  'Primary_Energy|',
                  'Price|',
                  'Production|']


    scenarioReplacementDict = {'historic'   :'Historic',
                               'Historical' : 'Historic',
                               'historical' : 'Historic',
                               'History'    : 'Historic',
                               'HISTCR'     : 'Historic|country_reported',
                               'HISTTP'     : 'Historic|third_party',
                               'computed historic' : 'Historic|computed'}
    
#inventory.category = None
#for entity in entityList:
#    mask = inventory.entity.str.startswith(entity)
#    
#    inventory.loc[mask, 'category'] = inventory.loc[mask, 'category'] + inventory.loc[mask, 'entity'].apply(lambda x: x.replace(entity,'')) 
#    inventory.loc[mask, 'entity'] = entity[:-1]
    outList = list()
    for table in tqdm.tqdm(tableSet):
#        table = tableSet[tableID]
        for string, newString in replacementDict.items():
            table.meta['entity'] = table.meta['entity'].replace(string,newString)
        
        for entity in entityList:
            if table.meta['entity'].startswith(entity):
                if 'category' in table.meta:
                    table.meta['category'] = '|'.join([table.meta['entity'].replace(entity,''), table.meta['category']]).lstrip('|')
                else:
                    table.meta['category'] = table.meta['entity'].replace(entity,'')
                table.meta['entity'] = entity.rstrip('|')
       
        for scenario in scenarioReplacementDict.keys():
            table.meta['scenario'] = table.meta['scenario'].replace(scenario, scenarioReplacementDict[scenario])
        
        if 'model' in table.meta.keys() and (table.meta['model'] in table.meta['scenario']):
            table.meta['scenario'] = table.meta['scenario'].replace(table.meta['model'],'').rstrip('|')
            table.generateTableID()
           
        sourceSplit = table.meta['source'].split('_') 
        if len(sourceSplit) ==2 :
            table.meta['source_name'], table.meta['source_year'] = sourceSplit
        else:
            if table.meta['source'].startswith('CAT'):
                table.meta['source_name'] = 'CAT'
                table.meta['source_year'] = table.meta['source'].replace('CAT_','')
            elif table.meta['source'].startswith('CA_NDCA'):
                table.meta['source_name'] = 'CA_NDCA'
                table.meta['source_year'] = table.meta['source'].replace('CA_NDCA_','')
            elif table.meta['source'].startswith('AIM_SSPx_DATA'):
                table.meta['source_name'] = 'AIM_SSPx_DATA'
                table.meta['source_year'] = table.meta['source'].replace('AIM_SSPx_DATA_','')
            elif table.meta['source'].startswith('CA_NDCA'):
                table.meta['source_name'] = 'CA_NDCA'
                table.meta['source_year'] = table.meta['source'].replace('CA_NDCA_','')
            elif table.meta['source'].startswith('IEA_CO2_FUEL'):
                table.meta['source_name'] = 'IEA_CO2_FUEL'
                table.meta['source_year'] = table.meta['source'].replace('IEA_CO2_FUEL_','')
            elif table.meta['source'].startswith('IEA_WEB'):
                table.meta['source_name'] = 'IEA_WEB'
                table.meta['source_year'] = table.meta['source'].replace('IEA_WEB_','')
            elif table.meta['source'].startswith('SDG_DB'):
                table.meta['source_name'] = 'SDG_DB'
                table.meta['source_year'] = table.meta['source'].replace('SDG_DB_','')
                  
            elif table.meta['source'].startswith('SSP_DB'):
                table.meta['source_name'] = 'SSP_DB'
                table.meta['source_year'] = table.meta['source'].replace('SSP_DB_','')
                  
            elif table.meta['source'].startswith('UNFCCC_CRF'):
                table.meta['source_name'] = 'UNFCCC_CRF'
                table.meta['source_year'] = table.meta['source'].replace('UNFCCC_CRF_','')
                  
            elif table.meta['source'].startswith('UN_WPP'):
                table.meta['source_name'] = 'UN_WPP'
                table.meta['source_year'] = table.meta['source'].replace('UN_WPP','')

                
        outList.append(table)
        
    return outList

#%%
def zipExport(IDList, fileName):
    from zipfile import ZipFile
    folder = os.path.join(config.PATH_TO_DATASHELF, 'exports/')
    os.makedirs(folder, exist_ok=True)
    
#    root = config.PATH_TO_DATASHELF
    
    sources = dt.find().loc[IDList].source.unique()
    sourceMeta = dt.core.DB.sources.loc[sources]
    sourceMeta.to_csv(os.path.join(folder, 'sources.csv'))
    zipObj = ZipFile(os.path.join(folder, fileName), 'w')
    zipObj.write(os.path.join(folder, 'sources.csv'),'./sources.csv')
    for ID in tqdm.tqdm(IDList):
        # Add multiple files to the zip
        tablePath = dt.core.DB._getPathOfTable(ID)
        csvFileName = os.path.basename(tablePath) 
        
        zipObj.write(tablePath,os.path.join('./data/', csvFileName))
#        zipObj.write(tablePath, os.path.relpath(os.path.join(root, file), os.path.join(tablePath, '..')))
 
    # close the Zip File
    zipObj.close()
    
def update_DB_from_zip_toV3(filePath, cleanTables=True):
    
#%%        
    from zipfile import ZipFile
    import shutil
    zf = ZipFile(filePath, 'r')
    
    tempFolder = os.path.join(config.PATH_TO_DATASHELF, 'temp/')
    shutil.rmtree(tempFolder, ignore_errors=True)
    os.makedirs(tempFolder)
    zf.extractall(tempFolder)
    zf.close()
    
    update_source_from_file(os.path.join(tempFolder, 'sources.csv'))
    tablesToUpdate = update_DB_from_folderV3(os.path.join(tempFolder, 'data'), message= 'DB update from ' + os.path.basename(filePath))
    return tablesToUpdate

def update_DB_from_zip(filePath):
    
#%%        
    from zipfile import ZipFile
    import shutil
    zf = ZipFile(filePath, 'r')
    
    tempFolder = os.path.join(config.PATH_TO_DATASHELF, 'temp/')
    shutil.rmtree(tempFolder, ignore_errors=True)
    os.makedirs(tempFolder)
    zf.extractall(tempFolder)
    zf.close()
    
    update_source_from_file(os.path.join(tempFolder, 'sources.csv'))
    update_DB_from_folder(os.path.join(tempFolder, 'data'), message= 'DB update from ' + os.path.basename(filePath))
#%%

def forAll(funcHandle, subset='scenario', source='IAMC15_2019_R2'):
    
    outTables = list()
    success = dict()
    if subset == "scenario":
        scenarios = dt.find(source=source).scenario.unique()
        
        for scenario in scenarios:
            try:
                outTables.append(funcHandle(scenario))
                print('{} run successfully'.format(scenario))
                success[scenario] = True
            except:
                #print('{} failed to run'.format(scenario))
                success[scenario] = False
                pass
    return outTables, success




import csv

def dict_to_csv(dictionary, filePath):

    with open(filePath, 'w', newline='') as file:
        writer = csv.writer(file)
        for key, val in dictionary.items():
            writer.writerow([key, val])
#%%
def csv_to_dict(filePath):

    with open(filePath, 'r', newline='') as file:
        reader = csv.reader(file)
        mydict = dict()
        for row in reader:
            print(row)
#            v = rows[1]
            mydict[row[0]] =  row[1]
    return mydict

def aggregate_table_to_region(table, mapping):
    #TODO remove soon
    raise(Warning('Depricated soon. Please use new implementaion in datatoolbox.tools.for_datatables'))
    
    missingCountryDict = dict()
    
    for region in mapping.listAll():

        
        missingCountries = set(mapping.membersOf(region)) - set(table.index)
#                print('missing countries: {}'.format(missingCountries))
        missingCountryDict[region] = list(missingCountries)
        availableCountries = set(mapping.membersOf(region)).intersection(table.index)
        if len(availableCountries) >0:
            table.loc[region,:] = table.loc[availableCountries,:].sum(axis=0, skipna=True)

    return table, missingCountryDict

def aggregate_tableset_to_region(tableSet, mapping):
    missingCountryDf = pd.DataFrame(columns=mapping.listAll())

    for tableKey in tableSet.keys():

        for region in mapping.listAll():
#                print(region)
            
            missingCountries = set(mapping.membersOf(region)) - set(tableSet[tableKey].index)
#                print('missing countries: {}'.format(missingCountries))
            missingCountryDf.loc[tableSet[tableKey].ID, region] = list(missingCountries)
            availableCountries = set(mapping.membersOf(region)).intersection(tableSet[tableKey].index)
            if len(availableCountries) >0:
                tableSet[tableKey].loc[region,:] = tableSet[tableKey].loc[availableCountries,:].sum(axis=0, skipna=True)

    return tableSet, missingCountryDf


import matplotlib.pylab as plt

def plotTables(tableList, countryList):
    fig = plt.figure(1)
    plt.clf()
    NUM_COLORS = len(countryList)
    
    cm = plt.get_cmap('gist_rainbow')
    coList = [cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)]
#    fig = plt.figure()
#    ax = fig.add_subplot(111)
#    ax.set_color_cycle([cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)])

#    ax.set_color_cycle([cm(1.*i/NUM_COLORS) for i in range(NUM_COLORS)])
    for table in tableList:
        for i, coISO in enumerate(countryList):
            plt.plot(table.columns, table.loc[coISO,:].T, color =coList[i] )
    plt.legend(countryList)
    plt.title(table.ID)


def pattern_match(
    data: pd.Series,
    patterns: Union[str, Iterable[str]],
    regex : bool = False
):
    """Find matches in `data` for a list of shell-style `patterns`

    Arguments
    ---------
    data : pd.Series
        Series of data to match against
    patterns : Union[str, Iterable[str]]
        One or multiple patterns, which are OR'd together
    regex : bool, optional
        Accept plain regex syntax instead of shell-style, default: False
    
    Returns
    -------
    matches : pd.Series
        Mask for selecting matched rows
    """

    if not isinstance(patterns, Iterable) or isinstance(patterns, str):
        patterns = [patterns]
    elif not patterns:
        raise ValueError("pattern list may not be empty")

    matches = False
    for pat in patterns:
        if isinstance(pat, str):
            if not regex:
                pat = shell_pattern_to_regex(pat) + '$'
            matches |= data.str.match(pat, na=False)
        else:
            matches |= data == pat

    return matches


def shell_pattern_to_regex(s):
    """Escape characters with specific regexp use"""
    return (
        str(s)
        .replace('|', r'\|')
        .replace('.', r'\.')  # `.` has to be replaced before `*`
        .replace('**', '__starstar__') # temporarily __starstar__
        .replace('*', r'[^|]*')
        .replace('__starstar__', r'.*')
        .replace('+', r'\+')
        .replace('(', r'\(')
        .replace(')', r'\)')
        .replace('$', r'\$')
    )

def fix_source_inconsistency(sourceID):
    gitManager = dt.core.DB.gitManager.__dict__['repositories'][sourceID]
    gitManager.git.execute(["git", "add", "tables/*"])
    dt.core.DB.gitManager.sources.loc['IEA_WEB_2020','git_commit_hash'] = gitManager.commit().hexsha
    dt.core.DB.gitManager.commit('inconsistent fix')


def _create_sandbox_tables(sourceID, random_seed):
    #%%
#    import datatoolbox as dt
    import numpy as np
    np.random.seed(1)
    

    tables = list()
    source_meta = {'SOURCE_ID': sourceID,
                      'collected_by' : 'Hard worker 1',
                      'date': core.getDateString(),
                      'source_url' : 'www.www.www',
                      'licence': 'open access' }
    
    meta = {'entity' : 'Emissions|CO2',
            'category': 'Total',
            'model': None,
            'scenario' : 'Historic',
            'source' : sourceID,
            'unit' : 'Mt CO2'}
    
    table = data_structures.Datatable(np.random.randint(0,20,[3,21]), 
                         columns = list(range(2000,2021)), 
                         index = ['World', 'Asia', 'ZAF'],
                         meta=meta).astype(float)
    tables.append(table)
    
    meta = {'entity' : 'Emissions|CO2',
            'category': 'Total',
            'scenario': 'Medium',
            'model' : 'Projection',
            'source' : sourceID,
            'unit' : 'Mt CO2'}
    
    table = data_structures.Datatable(np.random.randint(20,30,[3,31]), 
                         columns = list(range(2020,2051)), 
                         index = ['World', 'Asia', 'ZAF'],
                         meta=meta).astype(float)
    tables.append(table)
    
    
    
    meta = {'entity' : 'Emissions|CO2',
            'category': 'Total_excl_LULUCF',
            'scenario': None,
            'model' : 'Historic',
            'source_name' : sourceID,
            'unit' : 'Mt CO2'}
    
    table = data_structures.Datatable(np.random.randint(0,15,[3,21]), 
                         columns = list(range(2000,2021)), 
                         index = ['World', 'Asia', 'ZAF'],
                         meta=meta).astype(float)
    tables.append(table)
    return tables,source_meta


# old utilities


# open files
openers = dict()

operation_systems_supported = ['Linux', 'Darwin']
#excel
openers['xlsx'] = dict()
openers['xlsx']['Linux'] = 'libreoffice'
openers['xlsx']['Darwin'] = 'open -a "Microsoft Excel"'

openers['docx'] = dict()
openers['docx']['Linux'] = 'libreoffice'
openers['docx']['Darwin'] = 'open -a "Microsoft Word"'



def open_file(path):
    
    suffix = path.split('.')[-1]
    
    if config.OS not in operation_systems_supported:
        print('OS not supported. Currently support is restriced to ' + str(operation_systems_supported))
        return 
    
    if suffix not in openers.keys():
        print('no suiable file opener found ')
        return
    
    os.system(' '.join([openers[suffix][config.OS], 
                        path]))
        # elif dt.config.OS == 'Darwin':
        #     os.system('open -a "Microsoft Excel" ' + self.setup.MAPPING_FILE)


def shorten_find_output(dataframe):
    return dataframe.reset_index(drop=True).drop(['scenario', 'model', 'category', 'entity', 'source_year', 'source_name', 'unit'], axis=1)


def get_data_trees(**kwargs):
    findc = core.DB.findc
    results = findc(**kwargs)
    return _process_query(results)
    

def _process_query(results):
    
    # Initialize graphs for different data heads
    heads, graphs = [], {}
    
    if len(results.scenario.unique()) > 1:
        raise ValueError(
            "Multiple scenarios were detected, ensure that the"
            + " scenario name/ model/ data source etc. specify a unique scenario"
        )
    if len(results.scenario.unique()) == 0:
        raise ValueError(
            "Specified kwargs point to an empty list of scenarios. "
            + "Change kwargs or update your database."
        )
    scenario = list(results.scenario.unique())[0]

    # Iterate through data inventory
    for ix in results.index:

        # Remove scenario and model name
        ix = ix.split("__")[0]
        nodes = ix.split("|")

        # If first time occurrence of data head, update graphs
        if nodes[0] not in heads:
            heads.append(nodes[0])
            graph = nx.DiGraph()
            attr = {"label": nodes[0]}
            graph.add_node(nodes[0], **attr)
            graphs.update({nodes[0]: graph})

        # Fetch correct graph/  dict/ list
        graph = graphs[nodes[0]]

        # Add branches to tree
        root = nodes[0]
        for i, name_short in enumerate(nodes[1:]):
            # Get unique node name
            name_long = "|".join(nodes[1 : i + 2])
            # Add node to graph if it does not exist already
            if name_long not in graph.nodes:
                # Mark with "*"" if it point to a data table
                label = name_short
                # Add a short label for better visualization
                attr = {"label": label}
                graph.add_node(name_long, **attr)
            if i == len(nodes[1:]) - 1:
                graph.nodes[name_long]["label"] = name_short + "*"
            # Add edge
            graph.add_edge(root, name_long)
            root = name_long

    return graphs, scenario


def get_positions(graph, x_offset=1):
    """Get positions of nodes for horizontally aligned tree visualization

    Args:
        graph (networkx DiGraph): directed graph which is a tree
        x_offset (int, optional): Offset between horizontal spacing. Defaults to 1.

    Raises:
        TypeError: if not networkx DiGraph or tree

    Returns:
        dict: dictionary, mapping nodes to xy positions
    """

    # Check if tree
    if not isinstance(graph, nx.DiGraph):
        raise TypeError("Has to be a networkx DiGraph")

    if not nx.is_tree(graph):
        raise TypeError("Has to be a tree")

    # Determine root node
    root = next(iter(nx.topological_sort(graph)))

    # Determine number of subbranches
    out_degrees = graph.out_degree()

    def nb_leafs_in_subtree(root):
        """Recursive function for getting the number of leafs attached to root (root inclusive)

        Args:
            root (networkx node): root of subtree

        Returns:
            int: number of leafs in subtree
        """
        if out_degrees[root] == 0:
            nb_children = 1
        else:
            nb_children = sum(
                nb_leafs_in_subtree(child) for child in graph.neighbors(root)
            )

        return nb_children

    def set_positions(
        root_, x_spacing={}, depth=0, pos={root: (0, nb_leafs_in_subtree(root) / 2)},
    ):
        """Sets positions of nodes in a tree for horizontally aligned tree in a recursive fashion

        Args:
            root_ (networkx node): root of subtree
            x_spacing (dict, optional): Dictionary for keeping track of required horizontal spacing. Defaults to {}.
            depth (int, optional): Current tree depth. Defaults to 0.
            pos (dict, optional): [description]. Defaults to {root: (0, nb_leafs_in_subtree(root) / 2)}.

        Returns:
            (dict, dict): Returns  x_spacing and pos
        """

        # Consider length of root for x-spacing
        x_spacing.setdefault(depth, len(graph.nodes[root_]["label"]))
        x_spacing[depth] = max(x_spacing[depth], len(graph.nodes[root_]["label"]))

        if out_degrees[root_] == 0:
            return

        # Distribute children of root_ across the y-axis
        offset = 0
        depth += 1
        x_spacing.setdefault(depth, 0)

        for child in graph.neighbors(root_):
            y_pos = (
                pos[root_][1]
                - nb_leafs_in_subtree(root_) / 2
                + nb_leafs_in_subtree(child) / 2
                + offset
            )
            pos.update({child: (depth, y_pos)})
            offset += nb_leafs_in_subtree(child)

            set_positions(child, x_spacing, depth=depth, pos=pos)

        return pos, x_spacing

    # Determine positions of nodes
    pos, x_spacing = set_positions(root)
    # Re-adjust x-spacing
    pos = {
        key: (sum(x_spacing[i] + x_offset for i in range(pos_[0])), pos_[1])
        for key, pos_ in pos.items()
    }

    return pos, x_spacing


def plot_tree(
    graph,
    scenario,
    x_offset=3,
    fontsize=12,
    figsize=None,
    savefig_path=None,
    dpi=100,
):
    """Plots a tree indicating available data of a scenario 

    Parameters
    ----------
    graph : networkx.DiGraph
        tree in digraph format (obtained via get_data_trees function)
    scenario : str
        scenario name
    x_offset : int, optional
        x offset between nodes, by default 3
    fontsize : int, optional
        fontsize of the node labels (either fontsize or figsize can be specified not
        both), by default 12
    figsize : 2-dim tuple or None, optional
        figure size (either fontsize or figsize can be specified not
        both), by default None
    savefig_path : str or None, optional
        path to save figure to (e.g savefig_path = os.path.join(os.getcwd(), "fig.png") ),
        by default None
    dpi : int, optional
        dots per inches used in savefig, by default 100
    """

    pos, x_spacing = get_positions(graph, x_offset=x_offset)

    if figsize is None:
        fig, ax = plt.subplots()
    else:
        fig, ax = plt.subplots(figsize=figsize)

    # Draw the graph
    nx.draw(graph, pos=pos, with_labels=False, ax=ax, node_color="none")

    # Set xlim and ylim
    x_max = (
        sum(
            x + x_offset
            for level, x in x_spacing.items()
            if type(level) == int
        )
        - x_offset
    )
    y_max = max(pos_[1] for pos_ in pos.values()) + 1
    ax.set_xlim([0, x_max]), ax.set_ylim([0, y_max])

    # Get fontsize or reset figsize to avoid overlaps
    if fontsize is None:
        x_fig, y_fig = fig.get_size_inches() * fig.dpi
        fontsize = min(x_fig / (x_max / 1.5), y_fig / (y_max * 2.5))
    else:
        x_fig = 2 + (fontsize * (x_max / 1.5) / fig.dpi)
        y_fig = 2 + (fontsize * y_max * 2.5 / fig.dpi)
        fig.set_size_inches(x_fig, y_fig, forward=True)

    # Add node labels
    for node, xy in pos.items():
        text = graph.nodes[node]["label"]
        ax.annotate(
            text,
            xy,
            bbox=dict(pad=0.2, fc="gainsboro", ec="k", boxstyle="round"),
            family="monospace",
            fontsize=fontsize,
            verticalalignment="center",
            horizontalalignment="left",
        )

    # Add legend
    ax.annotate(
        "*: data available\nscenario: {}".format(scenario),
        (x_max, y_max),
        family="monospace",
        fontsize=fontsize,
        verticalalignment="top",
        horizontalalignment="right",
    )

    # Plot and save
    plt.tight_layout()

    if savefig_path is not None:
        plt.savefig(savefig_path, dpi=dpi)

    plt.show()

def plot_query_as_graph(results, savefig_path=None):
    
   graphs, scenario =  _process_query(results)
   for gKey in graphs.keys():
       plot_tree(graphs[gKey], 
                 scenario, 
#                 figsize=[5,6],
                 savefig_path=savefig_path)

def to_pyam(results, native_regions = False, disable_progress=None):
    """
    Load resuls as pyam IDateFrame.
    
    Parameters
    ----------
    results : pandas Dataframe with datatoolbox query results
        Results from find.
    native_regions : bool, optional
        Load native region defintions if available. The default is False.
    disable_progress : bool, optional
        Disable displaying of progressbar. The default None hides the
        progressbar on non-tty outputs.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    return core.DB.getTables(
        results.index, native_regions, disable_progress=disable_progress
    ).to_pyam()

def filterp(df, level=None, regex=False, **filters):
        """ 
        Future defaulf find method that allows for more
        sophisticated syntax in the filtering
        
        Usage:
        -------
        filters : Union[str, Iterable[str]]
            One or multiple patterns, which are OR'd together
        regex : bool, optional
            Accept plain regex syntax instead of shell-style, default: False
        
        Returns
        -------
        matches : pd.Series
        Mask for selecting matched rows
        """    
            
        # filter by columns and list of values
        keep = True

        for field, pattern in filters.items():
            # treat `col=None` as no filter applied
            if pattern is None:
                continue

            if field not in df:
                raise ValueError(f'filter by `{field}` not supported')

            keep &= pattern_match(
                df[field], pattern, regex=regex
            )

        if level is not None:
            keep &= df['variable'].str.count(r"\|") == level

        return df if keep is True else df.loc[keep]
def yearsColumnsOnly(index):
    """
    Extracts from any given index only the index list that can resemble 
    as year 
    
    e.g. 2001
    """
    
    import re
    REG_YEAR   = re.compile('^[0-9]{4}$')
    
    newColumns = []
    for col in index:
        if REG_YEAR.search(str(col)) is not None:
            newColumns.append(col)
        else:
            try: 
                if ~np.isnan(col) and REG_YEAR.search(str(int(col))) is not None:
                #   test float string
                    newColumns.append(col)
            except:
                pass
    return newColumns

def isin(df=None, **filters):
    """Constructs a MultiIndex selector

    Usage
    -----
    > df.loc[isin(region="World", gas=["CO2", "N2O"])]
    or with explicit df to get boolean mask
    > isin(df, region="World", gas=["CO2", "N2O"])
    """
    def tester(df):
        tests = (df.index.isin(np.atleast_1d(v), level=k) for k, v in filters.items())
        return reduce(and_, tests, next(tests))

    return tester if df is None else tester(df)

#%%    
if __name__ == '__main__':
    #%%
    def calculateTotalBiomass(scenario):
        source = 'IAMC15_2019_R2'
        tableID = core._createDatabaseID({"entity":"Primary_Energy|Biomass|Traditional",
                                         "category":"",
                                         "scenario":scenario,
                                         "source":'IAMC15_2019_R2'})
        tratBio = dt.getTable(tableID)
        
        tableID = core._createDatabaseID({"entity":"Primary_Energy|Biomass|Modern|wo_CCS",
                                         "category":"",
                                         "scenario":scenario,
                                         "source":'IAMC15_2019_R2'})
        modernBio = dt.getTable(tableID)
        
        tableID = core._createDatabaseID({"entity":"Primary_Energy|Biomass|Modern|w_CCS",
                                         "category":"",
                                         "scenario":scenario,
                                         "source":'IAMC15_2019_R2'})
        modernBioCCS = dt.getTable(tableID)
        
        table = tratBio + modernBio + modernBioCCS
        
        table.meta.update({"entity": "Primary_Energy|Biomass|Total",
                           "scenario" : scenario,
                           "source" : source,
                           "calculated" : "calculatedTotalBiomass.py",
                           "author" : 'AG'})
        return table

    outputTables, success = forAll(calculateTotalBiomass, "scenario")
    

def handle_exception(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.exception("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))   

def setup_logging(log_uncaught_exceptions=True, **kwargs):
    from colorlog import ColoredFormatter

    kwargs.setdefault("level", "INFO")

    streamhandler = logging.StreamHandler()
    streamhandler.setFormatter(
        ColoredFormatter(
            "%(name)-12s: %(log_color)s%(levelname)-8s%(reset)s %(message)s",
            datefmt=None,
            reset=True,
        )
    )

    kwargs.setdefault("handlers", []).append(streamhandler)
    
    if log_uncaught_exceptions:
        sys.excepthook = handle_exception

    logging.basicConfig(**kwargs)
    
    def add_parent_to_syspath():
        
        if '..' not in sys.path:
            sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 15 12:01:06 2020

@author: ageiges
"""
import os
import pandas as pd
import platform
import git
import os
OS = platform.system()
#from . import config
#from datatoolbox import config
import git
import shutil
import datatoolbox as dt

def create_empty_datashelf(pathToDataself,
                           force_new = False):
    """
    User funciton to create a empty datashelf

    Parameters
    ----------
    pathToDataself : TYPE
        DESCRIPTION.
    force_new : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    None.

    """
    from . import config 
    _create_empty_datashelf(pathToDataself, 
                           config.MODULE_PATH,
                           config.SOURCE_META_FIELDS,
                           config.INVENTORY_FIELDS,
                           force_new=force_new)

def compare_source_inventory_to_main(sourceID):
    git_manager = dt.database.GitRepository_Manager(dt.config,debugmode=True)
    inventory_csv_path = os.path.join(git_manager.PATH_TO_DATASHELF, 'inventory.csv')
    inventory = pd.read_csv(inventory_csv_path, index_col = 0)
      
    source_inventory_path = os.path.join(git_manager.PATH_TO_DATASHELF,  'database', sourceID, 'source_inventory.csv')
    source_inventory = pd.read_csv(source_inventory_path, index_col = 0)
    
    main_tables = inventory.index[inventory.source == sourceID]
    missing_tables_in_source = main_tables.difference(source_inventory.index)
    missing_tables_in_main = source_inventory.index.difference(main_tables)        
    if len(missing_tables_in_source) >0:
        print(f'{len(missing_tables_in_source)} are missing in source inventory')

    
    if len(missing_tables_in_main) >0:
        print(f'{len(missing_tables_in_main)} are missing in main inventory')
        
    if (len(missing_tables_in_main) == 0) and (len(missing_tables_in_source) == 0):
        print('Both inventories are equivalent')
        
    return missing_tables_in_source, missing_tables_in_main

def fix_uncommited_changes_in_source(sourceID):
    """
    Fix source when uncommited changes prevend the normal use of datatoolbox database.
    All uncommited local changes will be removed using "git reset --hard"

    Parameters
    ----------
    sourceID : str
        Source to be fixed.

    Returns
    -------
    None.

    """
    from . import config 
    git_manager = dt.database.GitRepository_Manager(config,debugmode=True)
    repo = git_manager.get_source_repo_failsave(sourceID)
    repo.git.reset('--hard','origin/master')
    print('All done')
    
def fix_inconsistent_source(sourceID, 
                            fix_what,
                            source_hash = None):
    """
    Fix inconsistent source if the datatoolbox sources csv does indicated a different
    git hash than the actual source after e.g. an manual commit.

    Parameters
    ----------
    sourceID : str
        Source to be fixed after a manual chagne
    fix_what : str
        What to fix - options are currently only "sources.csv".

    Returns
    -------
    None.

    """
    
    
    from . import config 
    git_manager = dt.database.GitRepository_Manager(config,debugmode=True)
    if fix_what == 'sources':
        main   = git_manager.repositories['main'] = git.Repo(git_manager.PATH_TO_DATASHELF)
        
        source = git.Repo(os.path.join(git_manager.PATH_TO_DATASHELF,  'database', sourceID))
        
        source_hash = source.head.object.hexsha
        
       
    elif fix_what == 'source_revision':
        main   = git_manager.repositories['main'] = git.Repo(git_manager.PATH_TO_DATASHELF)
        
        source_repo = git.Repo(os.path.join(git_manager.PATH_TO_DATASHELF,  'database', sourceID))
        source_repo.git.execute(["git", "reset", "--hard", source_hash])
        source_hash = source_repo.head.object.hexsha
        
        inventory_csv_path = os.path.join(git_manager.PATH_TO_DATASHELF, 'inventory.csv')
        inventory = pd.read_csv(inventory_csv_path, index_col = 0)
        
        ix_to_remove = inventory.index[inventory.source == sourceID]
        inventory = inventory.drop(ix_to_remove)
        
        source_inventory_path = os.path.join(git_manager.PATH_TO_DATASHELF,  'database', sourceID, 'source_inventory.csv')
        source_inventory = pd.read_csv(source_inventory_path, index_col = 0)
        
       
        inventory = pd.concat([inventory, source_inventory])
        inventory.to_csv(inventory_csv_path)
        main.index.add(inventory_csv_path)
        
    sources_csv_path = os.path.join(git_manager.PATH_TO_DATASHELF, 'sources.csv')
    
    sources_df = pd.read_csv(sources_csv_path, index_col =0)
    
    print (f'Overwriting has {sources_df.loc[sourceID, "git_commit_hash"]} -> {source_hash}')
    sources_df.loc[sourceID, 'git_commit_hash'] = source_hash
    sources_df.to_csv(sources_csv_path)

    main.index.add(sources_csv_path)
    main.index.commit(f'Fixed source {sourceID}' + " by " + config.CRUNCHER)
    
    
    
def get_source_log(sourceID = 'main',n =5 ):
    """
    Returns last log history of a source

    Parameters
    ----------
    sourceID : TYPE, optional
        DESCRIPTION. The default is 'main'.
    n : TYPE, optional
        DESCRIPTION. The default is 5.

    Returns
    -------
    None.

    """
    git_manager = dt.database.GitRepository_Manager(dt.config,debugmode=True)
    
    if sourceID == 'main':
        repo   = git.Repo(git_manager.PATH_TO_DATASHELF)
    else:
        repo   = git.Repo(os.path.join(git_manager.PATH_TO_DATASHELF,  'database', sourceID))
    # repo = git.Repo("/home/user/.emacs.d")
    commits = list(repo.iter_commits("master", max_count=n))
    print (f'============= Log of {sourceID} =============')
    for i in range(n):
        comm = commits[i]
        print(comm.committed_datetime.strftime('"%d-%m-%Y %H:%M') + ' - ' + comm.message + ' - \t Hash: ' + comm.hexsha)
    print (f'============= End log =============')
    
# def 
    
def _create_empty_datashelf(pathToDataself, 
                           MODULE_PATH,
                           SOURCE_META_FIELDS,
                           INVENTORY_FIELDS,
                           force_new=False):
    """
    Private function to create empty datashelf without propper 
    init of config
    """
    from pathlib import Path
    import os
    path = Path(pathToDataself)
    
    if force_new:
        shutil.rmtree(path)
    path.mkdir(parents=True,exist_ok=True)
    
    # add subfolders database
    Path(os.path.join(pathToDataself, 'database')).mkdir()
    Path(os.path.join(pathToDataself, 'mappings')).mkdir()
    Path(os.path.join(pathToDataself, 'rawdata')).mkdir()
    
    #create mappings
    os.makedirs(os.path.join(pathToDataself, 'mappings'),exist_ok=True)
    shutil.copyfile(os.path.join(MODULE_PATH, 'data/regions.csv'),
                    os.path.join(pathToDataself, 'mappings/regions.csv'))
    shutil.copyfile(os.path.join(MODULE_PATH, 'data/continent.csv'),
                    os.path.join(pathToDataself, 'mappings/continent.csv'))
    shutil.copyfile(os.path.join(MODULE_PATH, 'data/country_codes.csv'),
                    os.path.join(pathToDataself, 'mappings/country_codes.csv'))    
    
    sourcesDf = pd.DataFrame(columns = SOURCE_META_FIELDS)
    filePath= os.path.join(pathToDataself, 'sources.csv')
    sourcesDf.to_csv(filePath)
    
    inventoryDf = pd.DataFrame(columns = INVENTORY_FIELDS)
    filePath= os.path.join(pathToDataself, 'inventory.csv')
    inventoryDf.to_csv(filePath)
    git.Repo.init(pathToDataself)
        
def create_personal_setting(modulePath, OS):
    
    from . import config 
    os.makedirs(config.CONFIG_DIR, exist_ok=True)
    # Linux 
    if OS == 'Linux':
        import tkinter as tk
        from tkinter import simpledialog, filedialog
        
        
        ROOT = tk.Tk()
        ROOT.withdraw()
        userName = simpledialog.askstring(title="Initials",
                                          prompt="Enter your Initials or Name: ")
        print("Welcome", userName)


        root = tk.Tk()
        root.withdraw() #use to hide tkinter window
        
        def search_for_file_path ():
            currdir = os.getcwd()
            tempdir = filedialog.askdirectory(parent=root, initialdir=currdir, title='Please select a directory: ')
            if len(tempdir) > 0:
                print ("You chose: %s" % tempdir)
            return tempdir

        file_path_variable = search_for_file_path()
        
    else:
        userName = input("Please enter your initials: ")
        file_path_variable = input("Please enter path to datashelf: ")
    
    fin = open(os.path.join(modulePath, 'data','personal_template.py'), 'r')
    
    fout = open(os.path.join(config.CONFIG_DIR, 'personal.py'), 'w')
    
    for line in fin.readlines():
        outLine = line.replace('XX',userName).replace('/PPP/PPP', file_path_variable)
        fout.write(outLine)
    fin.close()
    fout.close()


def create_initial_config(module_path,
                          config_path, 
                          write_config=True):
    import git
    fin = open(os.path.join(module_path, 'data','personal_template.py'), 'r')
    
    
    
    DEBUG = False
    READ_ONLY = True
    sandboxPath = os.path.join(module_path, 'data','SANDBOX_datashelf')
    
    git.Repo.init(sandboxPath)
    
    if write_config:
        #create directory
        os.makedirs(config_path, exist_ok=True)
        fout = open(os.path.join(config_path, 'personal.py'), 'w')
        for line in fin.readlines():
            outLine = line.replace('/PPP/PPP', sandboxPath)
            fout.write(outLine)
        fout.close()
    fin.close()
    
    return 'XXX', sandboxPath, READ_ONLY, DEBUG

def change_personal_config():
#    from .tools.install_support import create_personal_setting
    modulePath =  os.path.dirname(__file__) + '/'
    create_personal_setting(modulePath, OS)
    
    

    
def _create_test_tables():
    """ 
    Creates tables in the database for testing
    """
    import numpy as np
    data = np.ones([4,5])


    ones = dt.Datatable(data, 
                  columns = [2010, 2012, 2013, 2015, 2016], 
                  index = ['ARG', 'DEU', 'FRA', 'GBR'],
                  meta={'entity' : 'Numbers',
                        'category' : 'Ones',
                       'scenario' : 'Historic',
                       'source' : 'Numbers_2020',
                       'unit' : 'm'} )
    
    fives = dt.Datatable(data*5, 
                  columns = [2010, 2012, 2013, 2015, 2016], 
                  index = ['ARG', 'DEU', 'FRA', 'GBR'],
                  meta={'entity' : 'Numbers',
                        'category' : 'Fives',
                       'scenario' : 'Historic',
                       'source' : 'Numbers_2020',
                       'unit' : 'm'} )

    sourceMeta = {'SOURCE_ID': 'Numbers_2020',
                 'collected_by': dt.config.CRUNCHER,
                 'date': '31.08.2020',
                 'source_url': 'datatoolbox',
                 'licence': 'free for all'}
    
    dt.commitTable(ones, 'add first table', sourceMeta)
    dt.commitTable(fives, 'add second table', sourceMeta)

def _re_link_functions(dt):

    
    dt.commitTable = dt.core.DB.commitTable
    dt.commitTables = dt.core.DB.commitTables
    
    dt.updateTable              = dt.core.DB.updateTable
    dt.updateTables             = dt.core.DB.updateTables
    dt.updateTablesAvailable    = dt.core.DB.updateTablesAvailable
    
    dt.removeTable = dt.core.DB.removeTable
    dt.removeTables = dt.core.DB.removeTables
    
    dt.findc            = dt.core.DB.findc
    dt.findp            = dt.core.DB.findp
    dt.find             = dt.core.DB.findc
    dt.findExact        = dt.core.DB.findExact
    dt.getTable         = dt.core.DB.getTable
    dt.getTables        = dt.core.DB.getTables
    dt.getTablesAvailable  = dt.core.DB.getTablesAvailable
    
    dt.isAvailable  = dt.core.DB._tableExists
    dt.validate_ID  = dt.core.DB.validate_ID
    dt.sourceInfo   = dt.core.DB.sourceInfo
    
    dt.updateExcelInput = dt.core.DB.updateExcelInput
    
def set_autoload_source(boolean):
    """
    Admin funtion to change permanently the personal configuration "AUTOLOAD_SOURCES".
    If set to True and if the database is connecte to a remote git repository,
    datatoolbox is try to import missing sources if a table is loaded that is locally 
    not available.

    Parameters
    ----------
    boolean : bool
        New config value.

    Returns
    -------
    None.

    """
    from . import config
    config.AUTOLOAD_SOURCES = boolean
    
    fin = open(os.path.join(config.CONFIG_DIR, 'personal.py'), 'r')
    lines = fin.readlines()
    fin.close()
    fout = open(os.path.join(config.CONFIG_DIR, 'personal.py'), 'w')
    
    line_found = False
    for line in lines:
        if line.startswith('AUTOLOAD'):
            outLine = 'AUTOLOAD_SOURCES = {}'.format(boolean)
            line_found = True
        else:
            outLine = line
            
        fout.write(outLine)
    
    if not line_found:
        # add it to old personal config
        outLine = 'AUTOLOAD_SOURCES = {}'.format(boolean)
        fout.write(outLine)
    fout.close()
    
    
def switch_database_to_testing():
    from . import config
#    from datatoolbox.tools.install_support import create_initial_config
#    
    _, sandboxPath, READ_ONLY, DEBUG = create_initial_config(config.MODULE_PATH, 
                                                             config_path=config.CONFIG_DIR,
                                                             write_config=False)
#
    if os.path.exists(os.path.join(sandboxPath, 'sources.csv')):
        shutil.rmtree(sandboxPath)
    
    create_empty_datashelf(sandboxPath)
    config.PATH_TO_DATASHELF = sandboxPath
    config.SOURCE_FILE = os.path.join(sandboxPath, 'sources.csv')
    dt.core.DB  = dt.database.Database()

    _re_link_functions(dt)
    _create_test_tables()
    
def naming_convention_test():
    #%%
    from naming_convention import entities
    
    
    variable_list = dt.findp().loc[:,['variable', 'entity']]
    entities_to_test = tuple(entities)
    
    inconsistent_ids = list()
    nan_ids          = list()
    for idx, (variable, entity) in variable_list.iterrows():
        if pd.isnull(entity):
            nan_ids.append(idx)
        else:
            if not idx.startswith(entities_to_test):
                inconsistent_ids.append(idx)
                
    #%%
    # finding entity
    tableList = list()
    for idx in nan_ids:
        
        variable, entity = variable_list.loc[idx,:]
        if idx.startswith(entities_to_test):
            candiates = list()
            
            for entity in entities_to_test:
                if idx.startswith(entity):
                    candiates.append(entity)
                    
                sort_candiates = sorted(candiates, key=len)
            
            entity = sort_candiates[-1]
            category = variable.lstrip(entity).lstrip('|')
            table = dt.getTable(idx)
            table.meta['category'] = category
            table.meta['entity'] = entity
            tableList.append(table)
    dt.updateTables([x.ID for x in tableList], tableList, 'added entity')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 09:54:20 2020

@author: ageiges
"""
#%% Import
import pandas as pd

#%%Defintions
_dataFrameStyle = """
<style>
    tr:nth-child(even) {color: blue; }
    tr:nth-child(odd)  {background: #ededed;}
    thead th      { background: #cecccc; }
    tr:hover { background: silver;   cursor: pointer;}
    td, th {padding: 5px;}
    {border: 1px solid silver;}
    {border-collapse: collapse; }
    {font-size: 11pt;}
    {font-family: Arial;}
    {text-align: left;}
</style>
"""
#nice green #d2f4e5

#%% Functions
def _HTML_with_style(df, style=None, random_id=None):
    """ 
    Generates HTML code based on a dataframe using a given styple format
    """
    
    from IPython.display import HTML
    import numpy as np
    import re

    df_html = df.to_html(escape =False)

    if random_id is None:
        random_id = 'id%d' % np.random.choice(np.arange(1000000))

    if style is None:
        style = """
        <style>
            table#{random_id} {{color: blue}}
        </style>
        """.format(random_id=random_id)
    else:
        new_style = []
        s = re.sub(r'</?style>', '', style).strip()
        for line in s.split('\n'):
                line = line.strip()
                if not re.match(r'^table', line):
                    line = re.sub(r'^', 'table ', line)
                new_style.append(line)
        new_style = ['<style>'] + new_style + ['</style>']

        style = re.sub(r'table(#\S+)?', 'table#%s' % random_id, '\n'.join(new_style))

    df_html = re.sub(r'<table', r'<table id=%s ' % random_id, df_html)

#    html_string = '''
#        <html>
#          <head><title>HTML Pandas Dataframe with CSS</title></head>
#          <link rel="stylesheet" type="text/css" href="df_style.css"/>
#          <body>
#          <h1>{heading}</h1>
#            {table}
#          </body>
#        </html>.
#        '''
    #df_html = html_string.format(table=df_html, heading=heading)
    return HTML(style + df_html)


def generate_html(dataframe, style =None) :
    """
    Generates basic html representation of a dataframe
    """
    
    if style is None:
        style = _dataFrameStyle
    pd.set_option('colheader_justify', 'center')   # FOR TABLE <th>
    
    html = _HTML_with_style(dataframe, '<style>table {}</style>'.format(style))
    
    return html.data.replace('NaN','')


def export_to_html( dataframe, fileName = None, heading =''):
    pd.set_option('colheader_justify', 'center')   # FOR TABLE <th>
    
    html = _HTML_with_style(dataframe, '<style>table {}</style>'.format(DataFrameStyle))
    
    html_string = '''
            <html>
              <head><title>HTML Pandas Dataframe with CSS</title></head>
              <link rel="stylesheet" type="text/css" href="df_style.css"/>
              <body>
              <h1>{heading}</h1>
                {table}
              </body>
            </html>
            '''
    df_html = html_string.format(table=html.data, heading=heading)
    
    if fileName:
        with open(fileName, 'w') as f:
            f.write(df_html)
    
    return df_html

#%% Classes
class HTML_BLOCK_FILE():
    MAIN_STYLE_HEADER  = """
                <style type="text/css">
                .main-container {
                  max-width: 940px;
                  margin-left: auto;
                  margin-right: auto;
                  text-align: justify;
                }
                h1 {
                    margin-top: 40px;
                }
                code {
                  color: inherit;
                  background-color: rgba(0, 0, 0, 0.04);
                }
                img {
                  max-width:100%;
                  height: auto;
                }
                .tabbed-pane {
                  padding-top: 12px;
                }
                .html-widget {
                  margin-bottom: 20px;
                }
                button.code-folding-btn:focus {
                  outline: none;
                }
                </style>
                <div class="container-fluid main-container">
                """

    MAIN_STYLE_FOODER  = """ </div>"""

    def __init__(self, fileName, title):
        
        self.fileName = fileName
        self.headerBlocks = list()
        self.bodyBlocks   = list()
        self.fooderBlocks  = list()
        self.title = title
        self.replaceDict = dict()
        
        self.headerBlocks.append(self.MAIN_STYLE_HEADER)
        self.fooderBlocks.append(self.MAIN_STYLE_FOODER)
        
    def _writeFile(self):
        buffer = self._createBuffer()
        buffer = self._doReplacement(buffer)
        with open(self.fileName, 'w') as f:
            f.write(buffer)
            
    def write(self):
        self._writeFile()
        
    def addToHeader(self, htmlCode):
        self.headerBlocks.append(htmlCode)
    
    def addToBody(self, htmlCode):
        self.bodyBlocks.append(htmlCode)
    
    def addToFooder(self, htmlCode):
        self.fooderBlocks.append(htmlCode)

    def appendHeading(self, heading, tierStr='h1'):
        self.bodyBlocks.append("""<{tier}>{heading}</{tier}>""".format(heading=heading, tier=tierStr))
        

    def addRepacement(self, old, new):
        self.replaceDict[old] = new

    def _doReplacement(self, buffer):
        for key, value in self.replaceDict.items():
            buffer = buffer.replace(key,value)
        return buffer
    
    def _createBuffer(self):
        #%% header 
        buffer = """<html>
                <head><title>{title}</title></head>
                <link rel="stylesheet" type="text/css" href="df_style.css"/>
                """.format(title=self.title)

        for block in self.headerBlocks:
            buffer += block
            
        #body
        buffer += """<body>"""

        for block in self.bodyBlocks:
            buffer += block
            
        buffer += """</body>"""
        
        #fooder 
        for block in self.fooderBlocks:
            buffer += block        
        buffer +="""</html>"""
        return buffer
        
class HTML_File():
    
    def __init__(self, fileName):
        
        self.fileName = fileName
        self.buffer = """"""
        

        self.mainStyle_header  = """
                <style type="text/css">
                .main-container {
                  max-width: 940px;
                  margin-left: auto;
                  margin-right: auto;
                  text-align: justify;
                }
                
                code {
                  color: inherit;
                  background-color: rgba(0, 0, 0, 0.04);
                }
                img {
                  max-width:100%;
                  height: auto;
                }
                .tabbed-pane {
                  padding-top: 12px;
                }
                .html-widget {
                  margin-bottom: 20px;
                }
                button.code-folding-btn:focus {
                  outline: none;
                }
                </style>
                <div class="container-fluid main-container">
                """


        self.mainStyle_fooder  = """ </div>"""

        self._appendHeader()
        
        
    def _appendHeader(self):
        self.buffer += """
        <html>
        """ + self.mainStyle_header + """
        <head><title>HTML Pandas Dataframe with CSS</title></head>
          <link rel="stylesheet" type="text/css" href="df_style.css"/>
          <body>
        """
    
    def _appendFooder(self):
        self.buffer += self.mainStyle_fooder + """
        </body>
        </html>.
        """
        
    def _writeFile(self):
        with open(self.fileName, 'w') as f:
            f.write(self.buffer)
            

    def appendHeading(self, heading, tierStr='h1'):
        self.buffer += """<{tier}>{heading}</{tier}>""".format(heading=heading, tier=tierStr)
        
    def appendTable(self, htmlTable):
        self.buffer +=htmlTable
    
    def appendText(self, htmlText):
        self.buffer += htmlText
        
        
    def close(self):
        self._appendFooder()
        self._writeFile()
        
